@@ -4,7 +4,7 @@
 #'
 #' @param dt A data frame with both x (predictor/feature) and y (response/label) variables.
 #' @param y Name of y variable.
-#' @param x Name of x variables. Default is NULL. If x is NULL, then all variables except y are counted as x variables.
+#' @param x Name of x variables. Default is NULL. If x is NULL, then all columns except y are counted as x variables.
 #' @param iv_limit The information value of kept variables should >= iv_limit. The default is 0.02.
 #' @param missing_limit The missing rate of kept variables should <= missing_limit. The default is 0.95.
 #' @param identical_limit The identical value rate (excluding NAs) of kept variables should <= identical_limit. The default is 0.95.
@@ -12,7 +12,8 @@
 #' @param var_kp Name of force kept variables, default is NULL.
 #' @param return_rm_reason Logical, default is FALSE.
 #' @param positive Value of positive class, default is "bad|1".
-#' @return A data.table with y and selected x variables and a data.table with the reason of removed x variable if return_rm_reason == TRUE.
+#'
+#' @return A dataframe with columns for y and selected x variables, and a dataframe with columns for remove reason if return_rm_reason == TRUE.
 #'
 #' @examples
 #' # Load German credit data
@@ -32,21 +33,23 @@
 #' @import data.table
 #' @export
 #'
-var_filter = function(dt, y, x = NULL, iv_limit = 0.02, missing_limit = 0.95, identical_limit = 0.95, var_rm = NULL, var_kp = NULL, return_rm_reason = FALSE, positive="bad|1") {
+var_filter = function(dt, y, x = NULL, iv_limit = 0.02, missing_limit = 0.95, identical_limit = 0.95, var_rm = NULL, var_kp = NULL, return_rm_reason = FALSE, positive = "bad|1") {
   # start time
   start_time = proc.time()
+  cat('[INFO] filtering variables ... \n')
 
   # no visible binding for global variable
   . = info_value = variable = rt = rm_reason = NULL
 
   # set dt as data.table
-  dt = setDT(dt)
-  # remove date/time col
-  dt = rmcol_datetime_unique1(dt)
-  # replace "" by NA
-  dt = rep_blank_na(dt)
+  dt = copy(setDT(dt))
+  if (!is.null(x)) dt = dt[, c(y,x), with=FALSE]
   # check y
   dt = check_y(dt, y, positive)
+  # # remove date/time col
+  # dt = rmcol_datetime_unique1(dt)
+  # # replace "" by NA
+  # dt = rep_blank_na(dt)
   # x variable names
   x = x_variable(dt,y,x)
 
